# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Extended set of commands for local dataset operations.

**NOTE**

Actually, the above isn't true at the moment. But that is the plan:
<https://github.com/datalad/datalad/issues/3192>
"""

__docformat__ = 'restructuredtext'
